/*
 *  GETFILES.C	    V1.30
 *
 *  DNET (c)Copyright 1988, Matthew Dillon, All Rights Reserved.
 *
 *
 *  GETFILES [-Nnetid] [-dlocaldir] [-c] file/dir file/dir file/dir
 *
 *  -Nnetid	network identifier
 *
 *  -dlocaldir	local directory to place files
 *
 *  -c		Continue from where you left off before.  Files that already
 *		exist on the local machine will not be re-transfered.  This
 *		will also re-start in the middle of a file that was
 *		partially transfered previously.
 *
 *		This command assumes the file(s) on both machines have not
 *		been modified since the first attempt.	No other checking
 *		is done at the moment.
 */

#include "defs.h"


typedef struct {
    char    Cmd;
    char    Str[128];
    long    Val;
} HDR;

#ifdef DEBUG
#define IFDEBUG(foo) foo
#else
#define IFDEBUG(foo) 
#endif


#ifndef __GNUC__
extern int Enable_Abort;
#endif

char Buf[4096];

short ContMode;
char  *NetId;
long  LocalDirLock;
void *Chan;

void fail ARGS((int));
int CheckNoPath ARGS((char *));
int LostChannel ARGS((int));
void UnknownCmd ARGS((HDR *));
int GetFile ARGS((HDR *, int));
int GetDir ARGS((HDR *, int));
int WriteHeader ARGS((char, char *, long));
int ReadHeader ARGS((HDR *));
char *NSpaces ARGS((int));
int CheckBreak ARGS((void));
void ResetBreak ARGS((void));

#ifdef MEM_DEBUG
void AllFree();
#endif
int brk()
{
    return(0);
}


main(ac,av)
int ac;
char *av[];
{
    HDR Hdr;
    short error;

    onbreak(brk);
#if !defined( LATTICE) && !defined(__GNUC__)
    Enable_Abort = 0;
#endif
#ifdef MEM_DEBUG
    atexit(AllFree);
#endif
    printf("GetFiles V%s%s\n", VERSION, GETFILES_VERSION);
    ResetBreak();
    IFDEBUG(printf("Doing args, locking dir\n");sleep(1);)
    {
	char *ldir = "";
	FIB *fib = malloc(sizeof(FIB));
	if(fib == NULL){
		fprintf(stderr, "Out of memory\n");
		exit(0);
	}
	ac = DoOption(ac, av, "N%sd%sc", &NetId, &ldir, &ContMode);
	IFDEBUG(printf("Options done\n");)
	if (ac <= 1) {
	    puts("GETFILES [-Nnetid -dlocaldir -c] remotefile/dir ...");
	    fail(22);
	}
	LocalDirLock = Lock(ldir, SHARED_LOCK);
	if (!LocalDirLock && ldir[0]) {     /*  couldn't find it, try to make it */
	    if (LocalDirLock = (long)CreateDir(ldir)) {
		UnLock(LocalDirLock);
		LocalDirLock = Lock(ldir, SHARED_LOCK);
	    }
	}
	if (!LocalDirLock || !Examine(LocalDirLock, fib) || fib->fib_DirEntryType < 0) {
	    printf("Unable to CD or create local directory: \"%s\"\n", ldir);
	    fail(21);
	}
	if(fib != NULL) free(fib);
    }
    IFDEBUG(printf("DOpen(%d)ing \n", PORT_GFILECOPY);sleep(1);)
    Chan = DOpen(NetId, PORT_GFILECOPY, 126, -80);
    if (!Chan) {
	puts("Unable to connect");
	fail(20);
    }
    IFDEBUG(printf("Net opened, saying hello\n");sleep(1);)
    error = WriteHeader('H', "Hello, getfiles client V1.30", 0);
    if (error)
	fail(LostChannel(1));
    IFDEBUG(printf("WroteHeader, Reading reply...\n");sleep(1);)
    switch(ReadHeader(&Hdr)) {
    case -1:
	fail(LostChannel(2));
    case 'H':
	printf("%s\n", Hdr.Str);
	break;
    }
    IFDEBUG(printf("ReadHeader, got Hello...\n");sleep(1);)
    {
	short i;

	for (i = 1; i < ac; ++i) {
	    short error;

            IFDEBUG(printf("WritingHdr(G, %s)\n", av[i]);sleep(1);)

	    error = WriteHeader('G', av[i], 0);
	    if (error)
		fail(LostChannel(3));
	    switch(ReadHeader(&Hdr)) {
	    case -1:
		fail(LostChannel(4));
	    case 'N':
		printf("Remote error on %s: %s\n", av[i], Hdr.Str);
		break;
	    case 'F':
		error = CheckNoPath(Hdr.Str);
		if (!error) {
		    long olddir = (long)CurrentDir(LocalDirLock);
		    IFDEBUG(printf("Getting file...\n");)
		    error = GetFile(&Hdr, 0);
		    IFDEBUG(printf("Going back to old dir...\n");sleep(1);)
		    CurrentDir(olddir);
		}
		break;
	    case 'D':
		error = CheckNoPath(Hdr.Str);
		if (!error) {
		    long olddir = (long)CurrentDir(LocalDirLock);
		    error = GetDir(&Hdr, 0);
		    CurrentDir(olddir);
		}
		break;
	    case 'S':
		printf("Access Violation: %s\n", Hdr.Str);
		break;
	    default:
		UnknownCmd(&Hdr);
		error = 1;
		break;
	    }
            IFDEBUG(printf("Done with that, err: %d\n", error);)
	    if (error)
		fail(error);
	}
	IFDEBUG(printf("Done with args, saying bye..\n");)
	if (!error) {
	    error = WriteHeader('E', "bye", 0);
	    if (error)
		fail(LostChannel(5));
	}
	IFDEBUG(printf("Done with everything, closing..\n");)
    }

    IFDEBUG(printf("Done closing, bye!\n");)
    fail(0);
}

void
fail(int code)
{
    IFDEBUG(printf("failing, code %d\n", code);sleep(1);)
    IFDEBUG(if(code == 0) code++;)

    if (Chan){
	if(code){printf("Attempting to close DNet channel...\n"); sleep(1);}
	DClose(Chan);
	if(code){printf("Whew, that worked...\n");sleep(1);}
    }
    if (LocalDirLock){
	if(code){printf("UnLock()ing LocalDir...\n");sleep(1);}
	UnLock(LocalDirLock);
	if(code){printf("Whew, that worked too.\n");sleep(1);}
    }

    if(code)printf("OK, bye\n");

    exit(code);   /* AllFree gets called via atexit */
}

int
CheckNoPath(str)
char *str;
{
    while (*str) {
	if (*str == '/' || *str == ':') {
	    puts("SECURITY ALERT: Illegal path spec received");
	    return(40);
	}
	++str;
    }
    return(0);
}

int
LostChannel(int point)
{
    printf("DATA CHANNEL LOST: %d\n", point);
    return(10);
}

void
UnknownCmd(hdr)
HDR *hdr;
{
    printf("UnRecognized command code: %02x\n", hdr->Cmd);
}

/*
 *  retrieve a file.  If ContMode set and file exists, try to append to
 *  it.
 */

int
GetFile(hdr, stab)
HDR *hdr;
int stab;
{
    BPTR Fh = 0;
    long pos = 0;
    short error = 0;

    IFDEBUG(printf("GetFile( %s, %d)\n", hdr->Str, stab);)

    printf("%s%-20s ", NSpaces(stab), hdr->Str);
    fflush(stdout);
    if (ContMode) {
	IFDEBUG(printf("ContMode\n");)
	SetProtection(hdr->Str, 0);         /*  make sure it's r/w  */
	if ((Fh = Open(hdr->Str, MODE_OLDFILE)) != 0) { /*  already exists  */
	    long len;
	    Seek(Fh, 0L, 1);
	    len = Seek(Fh, 0L, 0);          /*  get length      */
	    if (len > hdr->Val) {
		Close(Fh);
		printf("Cont Error, local file is larger than remote!: %s\n", hdr->Str);
		puts("(not downloaded)");
		return(0);
	    }
	    if (len == hdr->Val) {
		Close(Fh);
		if (error = WriteHeader('S', NULL, 0))
		    return(LostChannel(6));
		puts("HAVE IT, SKIP");
		return(0);
	    }
	    printf("(Append %ld/%ld) ", len, hdr->Val);
	    hdr->Val -= len;		/*  that much less  */
	    pos = len;			/*  start at offset */
	}
    }
    IFDEBUG(printf("Getting\n");)
    if (Fh == 0) {
	Fh = Open(hdr->Str, MODE_NEWFILE);
        IFDEBUG(printf("output file opened, handle %lx\n", Fh);)
	if (Fh == 0) {
	    error = WriteHeader('N', "open error", 0);
	    printf("Unable to open %s for output\n", hdr->Str);
	    if (error)
		return(LostChannel(7));
	    return(1);
	}
    }
    IFDEBUG(printf("WriteHeader(Y)\n");)
    error = WriteHeader('Y', NULL, pos);    /*  yes, gimme gimme    */

    /*
     *	Retrieve the data
     */

    IFDEBUG(printf("WriteHeader(y) result: %d\n", error);)
    if (error == 0) {
	long left = hdr->Val;
	long cnt = pos;
	long total = pos + left;
	IFDEBUG(printf("Receiving file\n");)

	while (left) {
	    long n = (left > sizeof(Buf)) ? sizeof(Buf) : left;
	    printf("%8ld/%-8ld\23317D", cnt, total);
	    fflush(stdout);
	    IFDEBUG(printf("Starting DRead \n");sleep(1);)
	    if (DRead(Chan, Buf, n) != n) {
		error = 5;
		break;
	    }
	    IFDEBUG(printf("Done DRead \n");)
	    if (CheckBreak()) {
		error = 1;
		break;
	    }
	    IFDEBUG(printf("Writing data\n");)
	    if (Write(Fh, Buf, n) != n) {
		puts("Local Write failed!");
		error = 6;
		break;
	    }
	    left -= n;
	    cnt += n;
	    if (CheckBreak()) {
		error = 1;
		break;
	    }
	    IFDEBUG(printf("Continuing receiving data\n");)
	}
	printf("%8ld/%-8ld  %s", cnt, total, ((cnt == total) ? "OK" : "INCOMPLETE"));
    }
    IFDEBUG(printf("Closing file (handle %lx)\n", Fh);sleep(1);)
    Close(Fh);
    IFDEBUG(printf("Closed\n");sleep(1);)
    puts("");
    if (error != 0) {
	SetProtection(hdr->Str, 016);    /*  disallow reads/write/exec! */
	IFDEBUG(printf("returning error\n");sleep(1);)
	return(LostChannel(8));
    }
    IFDEBUG(printf("Done getting \n");)
    return((int)error);
}

/*
 *  Retrieve a directory.  Create it if necessary.
 */

int
GetDir(hdr, stab)
HDR *hdr;
int stab;
{
    short error = 0;
    long dirlock;
    static HDR Hdr;	    /*	note: static */

    if (CheckBreak())
	return(1);
    printf("%s%-20s(DIR)\n", NSpaces(stab), hdr->Str);
    dirlock = Lock(hdr->Str, SHARED_LOCK);  /*  try to lock directory */
    if (!dirlock) {
	if (dirlock = (long)CreateDir(hdr->Str)) {
	    UnLock(dirlock);
	    dirlock = Lock(hdr->Str, SHARED_LOCK);
	}
    }
    if (!dirlock) {
	error = WriteHeader('N', "couldn't create", 0);
	printf("Unable to create local directory: %s\n", hdr->Str);
	if (error)
	    return(LostChannel(9));
	return(1);
    }
    error = WriteHeader('Y', NULL, 0);  /*  yes, gimme gimme    */
    while (!error) {
	switch(ReadHeader(&Hdr)) {
	case -1:
	    error = 1;
	    break;
	case 'E':                   /*  end of directory    */
	    UnLock(dirlock);
	    return(0);
	    break;
	case 'F':
	    error = CheckNoPath(Hdr.Str);
	    if (!error) {
		long olddir = (long)CurrentDir(dirlock);
		error = GetFile(&Hdr, stab + 4);
		CurrentDir(olddir);
	    }
	    break;
	case 'D':
	    error = CheckNoPath(Hdr.Str);
	    if (!error) {
		long olddir = (long)CurrentDir(dirlock);
		error = GetDir(&Hdr, stab + 4);
		CurrentDir(olddir);
	    }
	    break;
	case 'S':
	    printf("Access Violation: %s\n", Hdr.Str);
	    break;
	case 'N':
	    printf("REMOTE ERROR: %s\n", Hdr.Str);
	    error = 10;
	    break;
	default:
	    UnknownCmd(&Hdr);
	    error = 1;
	    break;
	}
    }
    UnLock(dirlock);
    return(LostChannel(10));
}

int
WriteHeader(c, str, len)
char c;
char *str;
long len;
{
    ubyte sl;
    char cmd = c;

    IFDEBUG(printf("Writing header %s \n", str);)

    if (str == NULL)
	str = "";
    sl = strlen(str);

    if (DWrite(Chan, &cmd, 1) < 0){
	return(1);
    }
    if (DWrite(Chan, &sl,1) < 0){
	return(2);
    }
    if (DWrite(Chan, str, sl) != sl){
	return(3);
    }
    if (DWrite(Chan, &len, sizeof(long) ) != 4){
	return(4);
    }
    return(0);
}

int
ReadHeader(hdr)
HDR *hdr;
{
    ubyte sl;
    ubyte cmd;

    hdr->Cmd = -1;
    if (DRead(Chan, &cmd, 1) != 1)
	return(-2);
    if (DRead(Chan, &sl, 1) != 1)
	return(-3);
    if (sl >= sizeof(hdr->Str)) {
	puts("Software error: received file name length too long");
	return(-4);
    }
    if (DRead(Chan, hdr->Str, sl) != sl)
	return(-5);
    hdr->Str[sl] = 0;
    if (DRead(Chan, &hdr->Val, sizeof(long) ) != 4)
	return(-6);
    hdr->Cmd = cmd;
    return((int)hdr->Cmd);
}

char *
NSpaces(n)
int n;
{
    static char Buf[128];
    static short in = 0;
    static short last;

    if (in == 0) {
	in = 1;
	BSet(Buf, sizeof(Buf)-1, ' ');
    }
    Buf[last] = ' ';
    if (n < 127)
	Buf[n] = 0;
    last = n;
    return(Buf);
}

int
CheckBreak()
{
    IFDEBUG(printf("SettingSignal\n");)
    if (SetSignal(0,0) & (SIGBREAKF_CTRL_C|SIGBREAKF_CTRL_D)) {
	puts("^C");
	IFDEBUG(printf("SettingSignal done 1\n");)
	return(1);
    }
    IFDEBUG(printf("SettingSignal done 0\n");)
    return(0);
}

void
ResetBreak()
{
    SetSignal(0, SIGBREAKF_CTRL_C|SIGBREAKF_CTRL_D);
}

